/*

    DGPS, Differential GPS Beacon Decoder
    Copyright (C) 1996  Richard Cochran

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/



/* 

Richard Cochran's DGPS beacon receiver software.

This software takes a raw syncronous bitstream as it
comes off the air from a USCG RTCM-104 DGPS beacon (usually
at 200 baud), and transforms it into a series of asyncronous 
bytes suitable for sending to a GPS receiver (normally 9600 baud).

The bitstream must be delivered to the computer via the CTS
handshake wire of a serial port.  The output to the GPS
is via the normal DATA OUT wire of the serial port, using
the standard UART.

The processing is divided up as follows:

 1. Read the 200 baud bitstream

	See baud200.cpp for details.  It has an init200() function
	which initializes an interrupt handler to watch the CTS handshake
	line, stuffing bits into a queue. and a getbit() function, which
	returns the next bit from the queue.

	Thanks to Kevin Schmidt, W9CF, for his program KS_CW
	(available with source code from your nearest SIMTEL archive).
	KS_CW is a program to decode CW and RTTY, and I modified its
	serial port routines for use in this DGPS software.  The serial
	port routines may have come from BYTE magazine originally.

 2. Send bits to the GPS.

	See the sendbittogps() function, below.  Basically, we just take
	6-bit chunks, flip the bits on the chunk (switch MSB and LSB),
	logically OR the chunk with 0x40, and send the resulting byte
	out the RS-232 port.  Note that, since RS-232 sends bytes LSB
	first, and the DGPS beacon sends bytes MSB first, when we
	"flip" the bits, we're really preserving the bit ordering in time.

 3. Sync up with the 200 baud bitstream

	See checkpar.cpp for details.  It has a loop which looks at
	30-bit words from the beacon, trying the GPS parity algorithm
	to synchronize itself with the data coming from the beacon.
	It also produces output giving information about the synchronization
	state.


If the PARITY_CHECK macro is #defined to be 0, we'll skip step 3, and
just send bytes to the GPS without syncing up or doing parity checking.
This works, but it's not recommended, as you get no feedback about how
well your receiver is performing.



*/



#define DGPS_EXTERN
#include "dgps.h"



void sendchartogps(int ch);
int flipbits(int);

void clear();
void init();
void end();
void usage();


int main (int argc, char * argv[])
{
	int i;
	// initialize default values
	com_port_num = COM_PORT_NUM;
	baud_rate = BAUD_RATE;
	beacon_baud_rate = BEACON_BAUD_RATE;

	for(i = 1; i < argc; i++) {
		if(argv[i][0] == '-' || argv[i][0] == '/') {
			switch(tolower(argv[i][1])) {
				case 'c': // com port num
					com_port_num = atoi(argv[i]+2);
					break;
				case 'b': // beacon baud rate
					beacon_baud_rate = atoi(argv[i]+2);
					if(beacon_baud_rate < 10 ||
					   beacon_baud_rate > 1000) {
						 printf("Invalid beacon baud rate, %d\n",
								beacon_baud_rate);
						 exit(1);
					}
					break;
				case 'g': // gps baud rate
					baud_rate = atoi(argv[i]+2);
					// make sure we have a valid baud rate
					switch (baud_rate) {
						case 300:
						case 600:
						case 1200:
						case 2400:
						case 4800:
						case 9600:
						case 19200:
						case 38400:
						case 57600:
						case 115200:
							break;  //these are ok, do nothing
						default:
							printf("Invalid GPS baud rate, %d\n",
									baud_rate);
							exit(1);
					} // end switch

					break;
				default:
					printf("Invalid argument - \"%s\"\n", argv[i]);
					usage();
			}
		} else {
			printf("Invalid argument - \"%s\"\n", argv[i]);
			usage();
		}
	}

	// set up serial port address and vector,
	// validating COM port number in the process.
	serial.address = portaddr(com_port_num);
	if(!serial.address){
		printf("Can't find com port, COM%d\n",
			   com_port_num);
		exit(1);
	}
	assert(serial.address == 0x3f8 ||
		   serial.address == 0x2f8 ||
		   serial.address == 0x3e8 ||
		   serial.address == 0x2e8);
	serial.vector = irqno(com_port_num);


	// start up the serial port and timer routines
	clear();
	init();

	end();
	return 0;
}

void usage() {
	printf(" valid options are -cX -bXXX -gXXXX\n");
	printf(" -cX     COM port number\n");

	exit(1);
}



void init()
{
	int gpsbauddivisor;

	// serial address better be valid now.
	assert(serial.address == 0x3f8 ||
		   serial.address == 0x2f8 ||
		   serial.address == 0x3e8 ||
		   serial.address == 0x2e8);

	// clear the abort flag, so we don't quit immediately
	abortflag = 0;

	// init 200 baud stuff
	init200();


	// set the baud rate
	bauddivisor = 115200 / baud_rate;
	outp(serial.address+3, 0x80);
	outpw(serial.address, bauddivisor);
	// set the LCR to N81.
	outp(serial.address+3, 3);

	// open dump file, if necessary
#if DUMPDATA
	datafile = fopen(DATAFILE, "w");
	assert(datafile);
#endif //DUMPDATA
}


void end()
{
	union REGS inregs, outregs;
	end200();
// clear the screen, unless we aborted in which case
// we want to see the final screen
	if(!abortflag) {
		inregs.h.ah = 0;
		inregs.h.al = video.mode;
		(void) int86(0x10,&inregs,&outregs);
	}
	_setcursortype(_NORMALCURSOR);

	// close dump file, if necessary
#if DUMPGPSDATA
	fclose (gpsdatafile);
#endif //DUMPGPSDATA

}

void sendchartogps(int ch) {
	while((inp(serial.address+5)&0x20)==0)
		; // port isn't ready, do nothing but wait
	outp(serial.address, ch);
}


// sendbittogps sends a single bit to the GPS.
// It queues up 6 bits into a byte and uses
// sendchartogps to send the byte off.
void sendbittogps(int bit) {
#if !(SEND_ONLY_WHEN_SYNCED)
	static int mod6count = 0;
	static int thisbyte = 0;
	assert(bit == 0 || bit == 1);

	thisbyte |= (bit << mod6count--);
	if(mod6count < 0){
		mod6count = 5;
		sendchartogps(flippedbits[thisbyte]);
		thisbyte = 0;
	}
#else
	bit++; // keeps compiler happy we're using parameter
#endif
}


// flipbits takes a 6-bit integer and flips the bit ordering
int flipbits(int in) {
	int i, rc;
	rc = 0;
	for(i = 0; i < 6; i++)
		rc |= (!!((0x20>>i) & in)) << i;
	return rc;
}

// sendwordtogps breaks up a 30-bit GPS word into 6-bit bytes
// and sends them to the GPS one at a time.
void sendwordtogps(long word) {

// unless SEND_ONLY_WHEN_SYNCED is defined, do nothing here --
// let sendbittogps handle it instead
#if SEND_ONLY_WHEN_SYNCED
	int i, thisbyte;
	for (i = 0; i < 5; i++) {
		thisbyte = flippedbits[(word & 0x3f000000) >> 24];
		sendchartogps(thisbyte);
		word <<= 6;
	}
#else
	word++; // keeps compiler happy that we're using the word
#endif


}

int irqno(int portno) {
	// This function automatically detects the IRQ number for a serial
	// port, given the serial port's number (1 for COM1, 2 for COM2, etc.)
	// Almost completely lifted from the serial port FAQ for University
	// of Illinois ECE 291.

	// returns: intlevel 0-15
	// if no intlevel found, issues error message and exits
	char ier,mcr,imrm,imrs,maskm,masks,irqm,irqs;
	unsigned base;

	base = portaddr(portno);
	if(!base)
		goto error_return;

	_asm cli;            // disable all CPU interrupts
	ier = inp(base+1);   // read IER
	outp(base+1,0);      // disable all UART ints
	while (!(inp(base+5)&0x20))
		;  // wait for the THR to be empty
	mcr = inp(base+4);   // read MCR
	outp(base+4,0x0F);   // connect UART to irq line
	imrm = inp(0x21);    // read contents of master ICU mask register
	imrs = inp(0xA1);    // read contents of slave ICU mask register
	outp(0xA0,0x0A);     // next read access to 0xA0 reads out IRR
	outp(0x20,0x0A);     // next read access to 0x20 reads out IRR
	outp(base+1,2);      // let's generate interrupts...
	maskm = inp(0x20);   // this clears all bits except for the one
	masks = inp(0xA0);   // that corresponds to the int
	outp(base+1,0);      // drop the int line
	maskm &= ~inp(0x20); // this clears all bits except for the one
	masks &= ~inp(0xA0); // that corresponds to the int
	outp(base+1,2);      // and raise it again just to be sure...
	maskm &= inp(0x20);  // this clears all bits except for the one
	masks &= inp(0xA0);  // that corresponds to the int
	outp(0xA1,~masks);   // now let us unmask this interrupt only
	outp(0x21,~maskm);
	outp(0xA0,0x0C);     // enter polled mode; Mike Surikov reported
	outp(0x20,0x0C);     // that order is important with Pentium/PCI systems
	irqs = inp(0xA0);    // and accept the interrupt
	irqm = inp(0x20);
	inp(base+2);         // reset transmitter interrupt in UART
	outp(base+4,mcr);    // restore old value of MCR
	outp(base+1,ier);    // restore old value of IER
	if (masks) outp(0xA0,0x20);  // send an EOI to slave
	if (maskm) outp(0x20,0x20);  // send an EOI to master
	outp(0x21,imrm);     // restore old mask register contents
	outp(0xA1,imrs);
	_asm sti;
	if (irqs&0x80)       // slave interrupt occured
		return (irqs&0x07)+8;
	if (irqm&0x80)       // master interrupt occured
		return irqm&0x07;

	error_return:
	printf("Can't find IRQ for COM%d.\n", portno);
	exit(1);
	return -1;
}


// portaddr returns the port address for a given COM port
// input: COM port number in range 1 - 4
// output: port address, 0 if invalid input or nonexistent port
int portaddr(int portno) {
	if(portno < 5 && portno > 0)
		return *((int far*)MK_FP(0x40, 2*(portno-1)));
	else
		return 0;
}



